
package org.mule.modules.twiml.config;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import org.mule.DefaultMuleEvent;
import org.mule.DefaultMuleMessage;
import org.mule.api.MessagingException;
import org.mule.api.MuleContext;
import org.mule.api.MuleEvent;
import org.mule.api.MuleException;
import org.mule.api.MuleMessage;
import org.mule.api.callback.HttpCallback;
import org.mule.api.construct.FlowConstruct;
import org.mule.api.construct.FlowConstructAware;
import org.mule.api.context.MuleContextAware;
import org.mule.api.expression.ExpressionManager;
import org.mule.api.lifecycle.Disposable;
import org.mule.api.lifecycle.Initialisable;
import org.mule.api.lifecycle.InitialisationException;
import org.mule.api.lifecycle.Startable;
import org.mule.api.lifecycle.Stoppable;
import org.mule.api.processor.MessageProcessor;
import org.mule.api.registry.RegistrationException;
import org.mule.api.transformer.DataType;
import org.mule.api.transformer.Transformer;
import org.mule.api.transformer.TransformerException;
import org.mule.config.i18n.CoreMessages;
import org.mule.config.i18n.MessageFactory;
import org.mule.construct.Flow;
import org.mule.modules.twiml.TwiMLModule;
import org.mule.modules.twiml.config.spring.DefaultHttpCallback;
import org.mule.modules.twiml.config.spring.NestedProcessorChain;
import org.mule.modules.twiml.config.spring.NestedProcessorString;
import org.mule.transformer.TransformerTemplate;
import org.mule.transformer.types.DataTypeFactory;
import org.mule.transport.NullPayload;
import org.mule.util.TemplateParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * DialMessageProcessor invokes the {@link org.mule.modules.twiml.TwiMLModule#dial(org.mule.api.callback.HttpCallback, int, boolean, int, java.lang.String, java.util.List)} method in {@link TwiMLModule }. For each argument there is a field in this processor to match it.  Before invoking the actual method the processor will evaluate and transform where possible to the expected argument type.
 * 
 */
public class DialMessageProcessor
    implements FlowConstructAware, MuleContextAware, Disposable, Initialisable, Startable, Stoppable, MessageProcessor
{

    /**
     * The flow to be invoked when the http callback is received
     * 
     */
    private Flow actionCallbackFlow;
    /**
     * An HttpCallback instance responsible for linking the APIs http callback with the flow {@link org.mule.modules.twiml.config.DialMessageProcessor#actionCallbackFlow
     * 
     */
    private HttpCallback action;
    private Object timeout;
    private int _timeoutType;
    private Object hangupOnStar;
    private boolean _hangupOnStarType;
    private Object timeLimit;
    private int _timeLimitType;
    private Object callerId;
    private String _callerIdType;
    private Object nestedProcessors;
    private List<org.mule.api.NestedProcessor> _nestedProcessorsType;
    private static Logger logger = LoggerFactory.getLogger(DialMessageProcessor.class);
    /**
     * Module object
     * 
     */
    private Object moduleObject;
    /**
     * Mule Context
     * 
     */
    private MuleContext muleContext;
    /**
     * Mule Expression Manager
     * 
     */
    private ExpressionManager expressionManager;
    /**
     * Mule Pattern Info
     * 
     */
    private TemplateParser.PatternInfo patternInfo;
    /**
     * Flow construct
     * 
     */
    private FlowConstruct flowConstruct;
    /**
     * Variable used to track how many retries we have attempted on this message processor
     * 
     */
    private AtomicInteger retryCount;
    /**
     * Maximum number of retries that can be attempted.
     * 
     */
    private int retryMax;

    /**
     * Obtains the expression manager from the Mule context and initialises the connector. If a target object  has not been set already it will search the Mule registry for a default one.
     * 
     * @throws InitialisationException
     */
    public void initialise()
        throws InitialisationException
    {
        retryCount = new AtomicInteger();
        expressionManager = muleContext.getExpressionManager();
        patternInfo = TemplateParser.createMuleStyleParser().getStyle();
        if (moduleObject == null) {
            try {
                moduleObject = muleContext.getRegistry().lookupObject(TwiMLModuleHttpCallbackAdapter.class);
                if (moduleObject == null) {
                    throw new InitialisationException(MessageFactory.createStaticMessage("Cannot find object"), this);
                }
            } catch (RegistrationException e) {
                throw new InitialisationException(CoreMessages.initialisationFailure("org.mule.modules.twiml.config.TwiMLModuleHttpCallbackAdapter"), e, this);
            }
        }
        if (nestedProcessors instanceof List) {
            for (MessageProcessor messageProcessor: ((List<MessageProcessor> ) nestedProcessors)) {
                if (messageProcessor instanceof Initialisable) {
                    ((Initialisable) messageProcessor).initialise();
                }
            }
        }
        if (actionCallbackFlow!= null) {
            action = new DefaultHttpCallback(actionCallbackFlow, muleContext, ((TwiMLModuleHttpCallbackAdapter) moduleObject).getDomain(), ((TwiMLModuleHttpCallbackAdapter) moduleObject).getLocalPort(), ((TwiMLModuleHttpCallbackAdapter) moduleObject).getRemotePort(), ((TwiMLModuleHttpCallbackAdapter) moduleObject).getAsync());
        }
    }

    public void start()
        throws MuleException
    {
        if (nestedProcessors instanceof List) {
            for (MessageProcessor messageProcessor: ((List<MessageProcessor> ) nestedProcessors)) {
                if (messageProcessor instanceof Startable) {
                    ((Startable) messageProcessor).start();
                }
            }
        }
        if (action!= null) {
            action.start();
        }
    }

    public void stop()
        throws MuleException
    {
        if (nestedProcessors instanceof List) {
            for (MessageProcessor messageProcessor: ((List<MessageProcessor> ) nestedProcessors)) {
                if (messageProcessor instanceof Stoppable) {
                    ((Stoppable) messageProcessor).stop();
                }
            }
        }
        if (action!= null) {
            action.stop();
        }
    }

    public void dispose() {
        if (nestedProcessors instanceof List) {
            for (MessageProcessor messageProcessor: ((List<MessageProcessor> ) nestedProcessors)) {
                if (messageProcessor instanceof Disposable) {
                    ((Disposable) messageProcessor).dispose();
                }
            }
        }
    }

    /**
     * Set the Mule context
     * 
     * @param context Mule context to set
     */
    public void setMuleContext(MuleContext context) {
        this.muleContext = context;
        if (nestedProcessors instanceof List) {
            for (MessageProcessor messageProcessor: ((List<MessageProcessor> ) nestedProcessors)) {
                if (messageProcessor instanceof MuleContextAware) {
                    ((MuleContextAware) messageProcessor).setMuleContext(context);
                }
            }
        }
    }

    /**
     * Sets flow construct
     * 
     * @param flowConstruct Flow construct to set
     */
    public void setFlowConstruct(FlowConstruct flowConstruct) {
        this.flowConstruct = flowConstruct;
        if (nestedProcessors instanceof List) {
            for (MessageProcessor messageProcessor: ((List<MessageProcessor> ) nestedProcessors)) {
                if (messageProcessor instanceof FlowConstructAware) {
                    ((FlowConstructAware) messageProcessor).setFlowConstruct(flowConstruct);
                }
            }
        }
    }

    /**
     * Sets the instance of the object under which the processor will execute
     * 
     * @param moduleObject Instace of the module
     */
    public void setModuleObject(Object moduleObject) {
        this.moduleObject = moduleObject;
    }

    /**
     * Sets retryMax
     * 
     * @param value Value to set
     */
    public void setRetryMax(int value) {
        this.retryMax = value;
    }

    /**
     * Sets timeLimit
     * 
     * @param value Value to set
     */
    public void setTimeLimit(Object value) {
        this.timeLimit = value;
    }

    /**
     * Sets hangupOnStar
     * 
     * @param value Value to set
     */
    public void setHangupOnStar(Object value) {
        this.hangupOnStar = value;
    }

    /**
     * Sets callerId
     * 
     * @param value Value to set
     */
    public void setCallerId(Object value) {
        this.callerId = value;
    }

    /**
     * Sets nestedProcessors
     * 
     * @param value Value to set
     */
    public void setNestedProcessors(Object value) {
        this.nestedProcessors = value;
    }

    /**
     * Sets actionCallbackFlow
     * 
     * @param value Value to set
     */
    public void setActionCallbackFlow(Flow value) {
        this.actionCallbackFlow = value;
    }

    /**
     * Sets timeout
     * 
     * @param value Value to set
     */
    public void setTimeout(Object value) {
        this.timeout = value;
    }

    /**
     * Get all superclasses and interfaces recursively.
     * 
     * @param classes List of classes to which to add all found super classes and interfaces.
     * @param clazz   The class to start the search with.
     */
    private void computeClassHierarchy(Class clazz, List classes) {
        for (Class current = clazz; (current!= null); current = current.getSuperclass()) {
            if (classes.contains(current)) {
                return ;
            }
            classes.add(current);
            for (Class currentInterface: current.getInterfaces()) {
                computeClassHierarchy(currentInterface, classes);
            }
        }
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link List }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isListClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(List.class);
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link Map }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isMapClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(Map.class);
    }

    private boolean isList(Type type) {
        if ((type instanceof Class)&&isListClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isList(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isList(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isMap(Type type) {
        if ((type instanceof Class)&&isMapClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isMap(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isMap(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isAssignableFrom(Type expectedType, Class clazz) {
        if (expectedType instanceof Class) {
            return ((Class) expectedType).isAssignableFrom(clazz);
        }
        if (expectedType instanceof ParameterizedType) {
            return isAssignableFrom(((ParameterizedType) expectedType).getRawType(), clazz);
        }
        if (expectedType instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) expectedType).getUpperBounds();
            if (upperBounds.length!= 0) {
                return isAssignableFrom(upperBounds[ 0 ], clazz);
            }
        }
        return false;
    }

    private Object evaluate(MuleMessage muleMessage, Object source) {
        if (source instanceof String) {
            String stringSource = ((String) source);
            if (stringSource.startsWith(patternInfo.getPrefix())&&stringSource.endsWith(patternInfo.getSuffix())) {
                return expressionManager.evaluate(stringSource, muleMessage);
            } else {
                return expressionManager.parse(stringSource, muleMessage);
            }
        }
        return source;
    }

    private Object evaluateAndTransform(MuleMessage muleMessage, Type expectedType, Object source)
        throws TransformerException
    {
        if (source == null) {
            return source;
        }
        Object target = null;
        if (isList(source.getClass())) {
            if (isList(expectedType)) {
                List newList = new ArrayList();
                Type valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                ListIterator iterator = ((List) source).listIterator();
                while (iterator.hasNext()) {
                    Object subTarget = iterator.next();
                    newList.add(evaluateAndTransform(muleMessage, valueType, subTarget));
                }
                target = newList;
            } else {
                target = source;
            }
        } else {
            if (isMap(source.getClass())) {
                if (isMap(expectedType)) {
                    Type keyType = Object.class;
                    Type valueType = Object.class;
                    if (expectedType instanceof ParameterizedType) {
                        keyType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                        valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 1 ];
                    }
                    Map map = ((Map) source);
                    Map newMap = new HashMap();
                    for (Object entryObj: map.entrySet()) {
                        {
                            Map.Entry entry = ((Map.Entry) entryObj);
                            Object newKey = evaluateAndTransform(muleMessage, keyType, entry.getKey());
                            Object newValue = evaluateAndTransform(muleMessage, valueType, entry.getValue());
                            newMap.put(newKey, newValue);
                        }
                    }
                    target = newMap;
                } else {
                    target = source;
                }
            } else {
                target = evaluate(muleMessage, source);
            }
        }
        if ((target!= null)&&(!isAssignableFrom(expectedType, target.getClass()))) {
            DataType sourceDataType = DataTypeFactory.create(target.getClass());
            DataType targetDataType = DataTypeFactory.create(((Class) expectedType));
            Transformer t = muleContext.getRegistry().lookupTransformer(sourceDataType, targetDataType);
            return t.transform(target);
        } else {
            return target;
        }
    }

    /**
     * Invokes the MessageProcessor.
     * 
     * @param event MuleEvent to be processed
     * @throws MuleException
     */
    public MuleEvent process(MuleEvent event)
        throws MuleException
    {
        MuleMessage muleMessage = event.getMessage();
        TwiMLModuleHttpCallbackAdapter castedModuleObject = null;
        if (moduleObject instanceof String) {
            castedModuleObject = ((TwiMLModuleHttpCallbackAdapter) muleContext.getRegistry().lookupObject(((String) moduleObject)));
            if (castedModuleObject == null) {
                throw new MessagingException(CoreMessages.failedToCreate("dial"), event, new RuntimeException("Cannot find the configuration specified by the config-ref attribute."));
            }
        } else {
            castedModuleObject = ((TwiMLModuleHttpCallbackAdapter) moduleObject);
        }
        try {
            Integer transformedTimeout = ((Integer) evaluateAndTransform(muleMessage, DialMessageProcessor.class.getDeclaredField("_timeoutType").getGenericType(), timeout));
            Boolean transformedHangupOnStar = ((Boolean) evaluateAndTransform(muleMessage, DialMessageProcessor.class.getDeclaredField("_hangupOnStarType").getGenericType(), hangupOnStar));
            Integer transformedTimeLimit = ((Integer) evaluateAndTransform(muleMessage, DialMessageProcessor.class.getDeclaredField("_timeLimitType").getGenericType(), timeLimit));
            String transformedCallerId = ((String) evaluateAndTransform(muleMessage, DialMessageProcessor.class.getDeclaredField("_callerIdType").getGenericType(), callerId));
            List<org.mule.api.NestedProcessor> transformedNestedProcessors = new ArrayList<org.mule.api.NestedProcessor>();
            if ((nestedProcessors!= null)&&(nestedProcessors instanceof List)) {
                for (MessageProcessor messageProcessor: ((List<MessageProcessor> ) nestedProcessors)) {
                    transformedNestedProcessors.add(new NestedProcessorChain(event, muleContext, messageProcessor));
                }
            } else {
                if ((nestedProcessors!= null)&&(nestedProcessors instanceof String)) {
                    transformedNestedProcessors.add(new NestedProcessorString(((String) nestedProcessors)));
                }
            }
            retryCount.getAndIncrement();
            Object resultPayload;
            resultPayload = castedModuleObject.dial(action, transformedTimeout, transformedHangupOnStar, transformedTimeLimit, transformedCallerId, transformedNestedProcessors);
            if (resultPayload == null) {
                event = new DefaultMuleEvent(new DefaultMuleMessage(NullPayload.getInstance(), muleContext), event);
            } else {
                List<Transformer> transformerList;
                transformerList = new ArrayList<Transformer>();
                transformerList.add(new TransformerTemplate(new TransformerTemplate.OverwitePayloadCallback(resultPayload)));
                event.getMessage().applyTransformers(event, transformerList);
            }
            retryCount.set(0);
            return event;
        } catch (Exception e) {
            throw new MessagingException(CoreMessages.failedToInvoke("dial"), event, e);
        }
    }

}
